﻿using System;
using GameKnight.Core;

namespace GameKnight
{
	
	namespace GameBoy
	{
	
		public sealed partial class Cartridge : ICartridge
		{
			
			public enum Types : byte
			{
				ROM = 0,
	        	MBC1 = 1,
	        	MBC1_RAM = 2,
	        	MBC1_RAM_BATT = 3,
	        	MBC2 = 5,
				MBC2_BATT = 6,
				ROM_RAM = 8,
				ROM_RAM_BATT = 9,
				MMM01 = 0x0B,
				MMM01_SRAM = 0x0C,
				MMM01_SRAM_BATT = 0x0D,
				MBC3_TIMER_BATT = 0x0F,
				MBC3_TIMER_RAM_BATT = 0x10,
				MBC3 = 0x11,
				MBC3_RAM = 0x12,
				MBC3_RAM_BATT = 0x13,
				MBC5 = 0x19,
				MBC5_RAM = 0x1A,
				MBC5_RAM_BATT = 0x1B,
				MBC5_RUMBLE = 0x1C,
				MBC5_RUMBLE_SRAM = 0x1D,
				MBC5_RUMBLE_SRAM_BATT = 0x1E,
				PocketCamera = 0x1F,
				TAMA5 = 0xFD,
				HuC3 = 0xFE,
				HuC1 = 0xFF,
			}

			public readonly bool ColorGameBoy = false;

			public readonly Bank[] RAM = new Bank[16];//null;
			
			public readonly Bank[] ROM = new Bank[512];//null;

            //public readonly bool SuperGameBoy = false;

			public readonly Types Type = Types.ROM;
				
			public Cartridge( IInterface Interface, IInputStream InputStream )
			{
				InputStream.Position = 0x104;
				if( ( InputStream.Length < 0x150 ) || ( InputStream.ReadUInt32() != 0x6666EDCE ) )
				{
					Interface.Error( this, "Not a valid GameBoy ROM file." );
					return;
				}
				InputStream.Position = 0x0143;
				this.ColorGameBoy = ( ( InputStream.ReadByte() & 0x80 ) != 0 );
				//BinaryReader.BaseStream.Position += 2;
				//this.SuperGameBoy = ( BinaryReader.ReadByte() == 0x03 );
				InputStream.Position = 0x0147;
				this.Type = (Types)InputStream.ReadByte();
				InputStream.Position = 0;
				this.ROM = new Bank[InputStream.Length>>14];
				for( int i = 0; i < ROM.Length; i++ )
				{
					this.ROM[i] = new Bank( InputStream.ReadBytes( 0x4000 ) );
				}
				switch( ROM[0][0x0149] )
				{
					case 1:
						RAM = new Bank[]{ new Bank( 0x800 ) };
						break;
					case 2:
						RAM = new Bank[1];
						break;
					case 3:
						RAM = new Bank[4];
						break;
					case 4:
					case 6:
						RAM = new Bank[16];
						break;
					case 5:
						RAM = new Bank[8];
						break;
				}  
				if( RAM != null )
				{
					for( int i = 0; i < RAM.Length; i++ )
					{
						if( RAM[i] == null )
						{
							RAM[i] = new Bank( 0x2000, 0xFF );
						}
					}
				}
				return;
			}	
			
			public Memory GetMapper( IConsole Console )
			{
				switch( Type )
				{
					case Types.ROM:
					case Types.ROM_RAM:
					case Types.ROM_RAM_BATT:
						return new MBC0( Console, this );
					case Cartridge.Types.MBC1:
					case Cartridge.Types.MBC1_RAM:
					case Cartridge.Types.MBC1_RAM_BATT:
						return new MBC1( Console, this );
					case Cartridge.Types.MBC2:
					case Cartridge.Types.MBC2_BATT:
						return new MBC2( Console, this );
					case Cartridge.Types.MBC3:
					case Cartridge.Types.MBC3_RAM:
					case Cartridge.Types.MBC3_RAM_BATT:
					case Cartridge.Types.MBC3_TIMER_BATT:
					case Cartridge.Types.MBC3_TIMER_RAM_BATT:
						return new MBC3( Console, this );
					case Cartridge.Types.MBC5:
					case Cartridge.Types.MBC5_RAM:
					case Cartridge.Types.MBC5_RAM_BATT:
					case Cartridge.Types.MBC5_RUMBLE:
					case Cartridge.Types.MBC5_RUMBLE_SRAM:
					case Cartridge.Types.MBC5_RUMBLE_SRAM_BATT:
						return new MBC5( Console, this );
				}
				Console.Interface.Error( this, "Unknown memory mapper, resorting to default." );
				return new MBC0( Console, this );
			}
			
		}
		
	}
	
}
