using System;
using GameKnight.Core;

namespace GameKnight
{
	
	namespace GameBoy
	{

		public sealed partial class Cartridge : ICartridge
		{
			
				public sealed class MBC3 : Memory
				{
	                 
					private sealed class TimerRegister
	            	{
	            		
						public int ClockControl = 0;
	
						public int ClockLatch = 0;
						
						public int ClockRegister = 0;				
	
	 					public int Days = 0;
						
						public int Hours = 0;
	
	  					public int LatchedControl = 0;
	
						public int LatchedDays = 0;
	
						public int LatchedHours = 0;
	
						public int LatchedMinutes = 0;
	
						public int LatchedSeconds = 0;
	
	  					public int Minutes = 0;
						
						public int Seconds = 0;
						
						public int TickCount = 0;
						
						public TimerRegister()
						{
							TickCount = System.Environment.TickCount;
							return;
						}
	
						public void Update()
						{
							int Now = System.Environment.TickCount;
							int Difference = Now - TickCount;
							if( Difference > 0 ) 
							{
							    Seconds += Difference % 60;
							    if( Seconds > 59 )
								{
									Seconds -= 60;
									Minutes++;
							    }
	    						Difference /= 60;
							    Minutes += Difference % 60;
							    if( Minutes > 60 )
								{
									Minutes -= 60;
									Hours++;
							    }
							    Difference /= 60;
								Hours += Difference % 24;
							    if( Hours > 24 )
								{
									Hours -= 24;
									Days++;
							    }
							    Difference /= 24;
							    Days += Difference;
							    if( Days > 255 )
								{
									if( Days > 511 )
									{
										Days %= 512;
										ClockControl |= 0x80;
									}
									ClockControl = ( ClockControl & 0xFE ) | ( ( Days>255 ) ? 1 : 0 );
								}
							}
							TickCount = Now;
							return;
						}
	
	            	}
	
					private TimerRegister Timer = new TimerRegister();
	
					public MBC3( IConsole Console, Cartridge Cartridge ) : base( Console, Cartridge )
					{
						return;
					}
					
					public override byte ReadU8( int Address ) 
					{
						switch( Address & 0xE000 )
						{
							case 0xA000:
								if( ( RAM != null ) && ( RAM_Enabled ) )
								{
									if( RAM_BankIndex >= 0 )
									{
										return RAM[RAM_BankIndex][Address];
									}
								}
								else
								{
									switch( Timer.ClockRegister )
									{
										case 0x08:
											return (byte)Timer.LatchedSeconds;
										case 0x09:
											return (byte)Timer.LatchedMinutes;
										case 0x0A:
											return (byte)Timer.LatchedHours;
										case 0x0B:
											return (byte)Timer.LatchedDays;
										case 0x0C:
											return (byte)Timer.LatchedControl;
									}
								}
								break;
							default:
                            	return base.ReadU8( Address );
						}
						return 0;
					}
	
					public override void WriteU8( int Address, byte Value )
					{
						switch( Address & 0xE000 )
						{
							case 0x0000:
								if( RAM != null )
								{
									RAM_Enabled = ( ( Value & 0x0F ) == 0x0A );
								}
								break;
							case 0x2000:
								ROM_BankIndex = ( Value & 0x7F );
								ROM_BankIndex %= ROM.Length;
								ROM_BankIndex = ( ( ROM_BankIndex == 0 ) ? 1 : ROM_BankIndex );
								break;
							case 0x4000:
								if( Value < 8 )
								{
									if( RAM != null )
									{
										RAM_BankIndex = ( Value % RAM.Length );
									}
	      						}
								else
								{
									if( RAM_Enabled )
									{
										RAM_BankIndex = -1;
										Timer.ClockRegister = Value;
									}
								}
								break;
							case 0x6000:
								if( ( Timer.ClockLatch == 0 ) && ( Value == 1 ) )
								{
									Timer.Update();
									Timer.LatchedSeconds = Timer.Seconds;
									Timer.LatchedMinutes = Timer.Minutes;
									Timer.LatchedHours = Timer.Hours;
									Timer.LatchedDays = Timer.Days;
									Timer.LatchedControl = Timer.ClockControl;
								}
								if( ( Value == 0 ) || ( Value == 1 ) )
								{
									Timer.ClockLatch = Value;
								}
								break;
							case 0xA000:
	                        	if( RAM_Enabled )
								{
									if( RAM_BankIndex >= 0 )
									{
										if( RAM != null )
										{
											RAM[RAM_BankIndex][Address] = Value;
										}
									}
									else
									{
										Timer.TickCount = System.Environment.TickCount;
										switch( Timer.ClockRegister )
										{
											case 0x08:
												Timer.Seconds = Value;
												break;
											case 0x09:
												Timer.Minutes = Value;
												break;
											case 0x0A:
												Timer.Hours = Value;
												break;
											case 0x0B:
												Timer.Days = Value;
												break;
											case 0x0C:
												if( ( Timer.ClockControl & 0x80 ) != 0 )
												{
													Timer.ClockControl = 0x80 | Value;
												}
												else
												{
													Timer.ClockControl = Value;
												}
												break;
										}
									}
								}
								break;
							default:
								base.WriteU8( Address, Value );
								break;
						}
						return;
					}	
	
				}
			
		}
		
	}
	
}


