﻿using System;
using GameKnight.Core;

namespace GameKnight
{
	
	namespace GameBoy
	{
	
		public abstract partial class Memory : IMemory, ISerializable
		{

			private readonly Register.BNK BNK = null;

			private readonly IConsole Console = null;
				
			//private readonly IIndexer[] IndexerArray = new IIndexer[256];
			
			protected Bank[] RAM = null;
				
			protected int RAM_BankIndex = 0;
	
			protected bool RAM_Enabled = false;
		
			private readonly IRegister[] RegisterArray = new IRegister[0x100];

			protected readonly Bank[] ROM = null;
						    
			protected int ROM_BankIndex = 1;
			
			private readonly ISprite[] SpriteArray = new ISprite[0x40];

            private readonly Bank[] WRAM = new Bank[8]{ new Bank( 0x1000 ), new Bank( 0x1000 ), new Bank( 0x1000 ), new Bank( 0x1000 ), new Bank( 0x1000 ), new Bank( 0x1000 ), new Bank( 0x1000 ), new Bank( 0x1000 ) };

			public Memory( IConsole Console, Cartridge Cartridge )
			{
				this.Console = Console;
				this.ROM = Cartridge.ROM;
				this.RAM = Cartridge.RAM;
				this.SpriteArray = Console.VideoMemory.GetSprites();
				for( int i = 0; i < RegisterArray.Length; i++ )
				{
					this.RegisterArray[i] = Console.GetRegister( 0xFF00 | i );
				}
				this.BNK = Console.GetRegister<Register.BNK>();
				return;
			}

			public void Copy( int Source, int Destination, int Count )
			{
				while( Count > 0 )
				{
					WriteU8( Destination++, ReadU8( Source++ ) );
					Count--;
				}
				return;
			}
			
			public sbyte ReadS8( int Address )
			{
				return (sbyte)ReadU8( Address );
			}
			
			public virtual byte ReadU8( int Address )
			{		
				switch( Address & 0xF000 )
				{
					case 0x0000:
					case 0x1000:
					case 0x2000:
					case 0x3000:
						return ROM[0][Address];
					case 0x4000:
					case 0x5000:
					case 0x6000:
					case 0x7000:
						return ROM[ROM_BankIndex][Address];
					case 0x8000:
					case 0x9000:
						if( BNK != null )
						{
							return Console.VideoMemory.ReadU8( ( BNK.VideoBank << 13 ) | Address );
						}
						return Console.VideoMemory.ReadU8( Address );
					case 0xA000:
					case 0xB000:
						if( ( RAM != null ) && ( RAM_Enabled ) )
						{
							return RAM[RAM_BankIndex][Address];
						}
						break;
					case 0xC000:
					case 0xE000:
						return WRAM[0][Address];
					case 0xD000:	
					case 0xF000:
						if( Address >= 0xFF00 )
						{
							return RegisterArray[Address&0xFF][Address];
						}
						else if( Address >= 0xFE00 )
						{		
                        	if( Address < 0xFEA0 )
							{
								return SpriteArray[((Address>>2)&0x3F)][Address];
							}
                        }
						else if( BNK != null )
						{
							return WRAM[BNK.WorkBank][Address];
						}
						else
						{
							return WRAM[1][Address];
						}
						break;
					default:
						break;
				}
				return 0;
			}

			public ushort ReadU16( int Address ) 
			{
				return (ushort)( ReadU8( Address ) | ( ReadU8( Address + 1 ) << 8 ) );
			}	
	
			public virtual void WriteU8( int Address, byte Value )
			{
				switch( Address & 0xF000 )
				{
					case 0x8000:
					case 0x9000:
						if( BNK != null )
						{
							Console.VideoMemory.WriteU8( ( ( BNK.VideoBank << 13 ) | Address ), Value );
						}
						else
						{
							Console.VideoMemory.WriteU8( Address, Value );
						}
						break;
					case 0xA000:
					case 0xB000:
						if( ( RAM != null ) && ( RAM_Enabled ) )
						{
							RAM[RAM_BankIndex][Address] = Value;
						}
						break;
					case 0xC000:
					case 0xE000:
						WRAM[0][Address] = Value;
						break;
					case 0xD000:
					case 0xF000:
						if( Address >= 0xFF00 )
						{
							RegisterArray[Address&0xFF][Address] = Value;
						}
						else if( Address >= 0xFE00 )
						{
							if( Address < 0xFEA0 )
							{
								SpriteArray[(Address>>2)&0x3F][Address] = Value;
							}
						}
						else if( BNK != null )
						{
							WRAM[BNK.WorkBank][Address] = Value;
						}
						else
						{
							WRAM[1][Address] = Value;
						}
						break;
					default:
						break;
				}
				return;
			}
			
			public void WriteU16( int Address, ushort Value ) 
			{
				WriteU8( Address, (byte)( Value & 0xFF ) );
				WriteU8( Address + 1, (byte)( Value >> 8 ) );
				return;
			}
			
		}
		
	}
	
}
