﻿using System;
using GameKnight.Core;

namespace GameKnight
{
	
	namespace GameBoy
	{
	
		public partial class Processor : IProcessor
		{
			
			protected partial class Opcode
			{
				
				public static int ADC( Processor Processor, Opcode Opcode )
				{
					byte Value = Processor.GetRegisterValue( ( Opcode.Value & 0x07 ), ( Opcode.Value >= 0xC0 ) );
					Temp.Value = (ushort)( Processor.A + Value + ( Processor.CarryFlag ? 1 : 0 ) );
					Processor.NegativeFlag = false;
					Processor.CarryFlag = ( Temp.UpperByte != 0 );
					Processor.HalfCarryFlag = ( ( ( Processor.A ^ Value ^ Temp.LowerByte ) & 0x10 ) != 0 );
					Processor.ZeroFlag = ( Temp.LowerByte == 0 );
					Processor.A = Temp.LowerByte;
					return 0;
				}

				public static int ADD( Processor Processor, Opcode Opcode )
				{
					byte Value = Processor.GetRegisterValue( ( Opcode.Value & 0x07 ), ( Opcode.Value >= 0xC0 ) );
					Temp.Value = (ushort)( Processor.A + Value );
					Processor.NegativeFlag = false;
					Processor.CarryFlag = ( Temp.UpperByte != 0 );
					Processor.HalfCarryFlag = ( ( ( Processor.A ^ Value ^ Temp.LowerByte ) & 0x10 ) != 0 );
					Processor.ZeroFlag = ( Temp.LowerByte == 0 );
					Processor.A = Temp.LowerByte;
					return 0;
				}
				
				public static int AND( Processor Processor, Opcode Opcode )
				{
					Processor.A &= Processor.GetRegisterValue( ( Opcode.Value & 0x07 ), ( Opcode.Value >= 0xC0 ) );
					Processor.ClearFlags();
					Processor.HalfCarryFlag = true;
					Processor.ZeroFlag = ( Processor.A == 0 );
					return 0;
				}
					
				public static int CP( Processor Processor, Opcode Opcode )
				{
					byte Value = Processor.GetRegisterValue( ( Opcode.Value & 0x07 ), ( Opcode.Value >= 0xC0 ) );
					Temp.Value = (ushort)( Processor.A - Value );
					Processor.NegativeFlag = true;
					Processor.CarryFlag = ( Temp.UpperByte != 0 );
					Processor.HalfCarryFlag = ( ( ( Processor.A ^ Value ^ Temp.LowerByte ) & 0x10 ) != 0 );
					Processor.ZeroFlag = ( Temp.LowerByte == 0 );
					return 0;
				}
				
				public static int DEC8( Processor Processor, Opcode Opcode )
				{
					int RegisterIndex = ( ( Opcode.Value >> 3 ) & 0x07 );
					byte Value = (byte)( Processor.GetRegisterValue( RegisterIndex ) - 1 );
					Processor.NegativeFlag = true;
					Processor.HalfCarryFlag = ( ( Value & 0x0F ) == 0x0F );
					Processor.ZeroFlag = ( Value == 0 );
					Processor.SetRegisterValue( RegisterIndex, Value );
					return 0;
				}

				public static int INC8( Processor Processor, Opcode Opcode )
				{
					int RegisterIndex = ( ( Opcode.Value >> 3 ) & 0x07 );
					byte Value = (byte)( Processor.GetRegisterValue( RegisterIndex ) + 1 );
					Processor.NegativeFlag = false;
					Processor.HalfCarryFlag = ( ( Value & 0x0F ) == 0 );
					Processor.ZeroFlag = ( Value == 0 );
					Processor.SetRegisterValue( RegisterIndex, Value );
					return 0;
				}
				
				public static int OR( Processor Processor, Opcode Opcode )
				{
					Processor.A |= Processor.GetRegisterValue( ( Opcode.Value & 0x07 ), ( Opcode.Value >= 0xC0 ) );
					Processor.ClearFlags();
					Processor.ZeroFlag = ( Processor.A == 0 );
					return 0;
				}
				
				public static int SBC( Processor Processor, Opcode Opcode )
				{
					byte Value = Processor.GetRegisterValue( ( Opcode.Value & 0x07 ), ( Opcode.Value >= 0xC0 ) );
					Temp.Value = (ushort)( Processor.A - Value - ( Processor.CarryFlag ? 1 : 0 ) );
					Processor.NegativeFlag = true;
					Processor.CarryFlag = ( Temp.UpperByte != 0 );
					Processor.HalfCarryFlag = ( ( ( Processor.A ^ Value ^ Temp.LowerByte ) & 0x10 ) != 0 );
					Processor.ZeroFlag = ( Temp.LowerByte == 0 );
					Processor.A = Temp.LowerByte;
					return 0;
				}

				public static int SUB( Processor Processor, Opcode Opcode )
				{
					byte Value = Processor.GetRegisterValue( ( Opcode.Value & 0x07 ), ( Opcode.Value >= 0xC0 ) );
					Temp.Value = (ushort)( Processor.A - Value );
					Processor.NegativeFlag = true;
					Processor.CarryFlag = ( Temp.UpperByte != 0 );
					Processor.HalfCarryFlag = ( ( ( Processor.A ^ Value ^ Temp.LowerByte ) & 0x10 ) != 0 );
					Processor.ZeroFlag = ( Temp.LowerByte == 0 );
					Processor.A = Temp.LowerByte;
					return 0;
				}

				public static int XOR( Processor Processor, Opcode Opcode )
				{
					Processor.A ^= Processor.GetRegisterValue( ( Opcode.Value & 0x07 ), ( Opcode.Value >= 0xC0 ) );
					Processor.ClearFlags();
					Processor.ZeroFlag = ( Processor.A == 0 );
					return 0;
				}

			}
			
		}
		
	}
	
}
