﻿using System;
using Text = System.Text;
using GameKnight.Core;

namespace GameKnight
{
	
	namespace GameBoy
	{
	
		public partial class Processor : IProcessor
		{
			
			protected partial class Opcode
			{
			
				[Flags]
				public enum AddressModes : int
				{
					None = 0,
					Immediate = 1,
					ImmediateExtended = 2,
				}   
				
				public enum Conditions : int
				{
					NZ = 0,
					Z = 1,
					NC = 2,
					C = 3,
				};
				
				private readonly AddressModes AddressMode = AddressModes.None;
				
				public readonly int Cycles = 0;
					
				public delegate int Delegate( Processor Processor, Opcode Opcode );
				
				public readonly Delegate Function = null;
				
				private readonly string Mnemonic = "";
				
				public int Size
				{
					get
					{
						return 1 + ( ( Value >= 0x100 ) ? 1 : 0 ) + ( ( ( AddressMode & AddressModes.Immediate ) != 0 ) ? 1 : 0 ) + ( ( ( AddressMode & AddressModes.ImmediateExtended ) != 0 ) ? 2 : 0 );
					}
				}
				
				private static RegisterPair Temp = new RegisterPair( 0 );

				public readonly int Value = 0;
				
				private Opcode( int Value, Delegate Function, int Cycles, string Mnemonic )
				{
					this.Value = Value;
					this.Function = Function;
					this.Cycles = Cycles;
					this.Mnemonic = Mnemonic;
					if( ( Mnemonic.Contains( "%n" ) ) || ( Mnemonic.Contains( "%e" ) ) || ( Mnemonic.Contains( "%o" ) ) )
					{
						AddressMode = AddressModes.Immediate;
					}
					else if( Mnemonic.Contains( "%N" ) )
					{
						AddressMode = AddressModes.ImmediateExtended;	
					}
					return;
				}
				
				public int Execute( Processor Processor )
				{
					Function( Processor, this );
					return Cycles;
				}
	
				public string ToString( Processor Processor, int Address )
				{
					Address += ( ( Value < 0x100 ) ? 1 : 2 );
					Text.StringBuilder StringBuilder = new Text.StringBuilder( Mnemonic );
					StringBuilder.Replace( "%n", String.Format( "{0:X2}h", Processor.Memory.ReadU8( Address ) ) );
					StringBuilder.Replace( "%o", String.Format( "{0:X2}h", Processor.Memory.ReadU8( Address ) ) );
					StringBuilder.Replace( "%e", String.Format( "{0:X4}h", Address + 1 + Processor.Memory.ReadS8( Address ) ) );
					StringBuilder.Replace( "%b", ( ( Value >> 3 ) & 0x07 ).ToString() );
					StringBuilder.Replace( "%t", String.Format( "{0:X2}h", ( Value & 0x38 ) ) );
					StringBuilder.Replace( "%r0", Processor.RegisterNameArray[(Value&0x07)] );
					StringBuilder.Replace( "%r3", Processor.RegisterNameArray[((Value>>3)&0x07)] );
					StringBuilder.Replace( "%N", String.Format( "{0:X4}h", Processor.Memory.ReadU16( Address ) ) );
					StringBuilder.Replace( "%R", Processor.RegisterPairNameArray[((Value>>4)&0x03)] );
					StringBuilder.Replace( "%P", Processor.RegisterPairNameArray[((((Value>>4)&0x03)!=3)?((Value>>4)&0x03):4)] );
					StringBuilder.Replace( "%C", ((Conditions)((Value>>3)&0x03)).ToString() );
					return StringBuilder.ToString();
				}
				
			}
			
		}
		
	}
	
}
