﻿using System;
using GameKnight.Core;

namespace GameKnight
{
	
	namespace GameBoy
	{

		public partial class Register : IRegister
		{
			
			public class HDMA : IRegister
			{

            	private readonly IConsole Console = null;

				private int Count = 0;

				private readonly byte[] DataArray = new byte[6] { 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF };

                private int Destination = 0x8000;

				public bool Enabled
				{
					get
					{
						return EnabledProperty;
					}
					private set
					{
                    	EnabledProperty = value;
					}
				}

				private bool EnabledProperty = false;

				public bool HaltFlag = false;

                private readonly Register.LCDC LCDC = null;

                private int Source = 0;

				public HDMA( IConsole Console )
				{
					this.Console = Console;
					this.LCDC = Console.GetRegister<Register.LCDC>();
					return;
				}
				
                public byte this[int Address]
				{
                	get
                	{
						return DataArray[Address&0x07];
                	}
                	set
                	{
                		switch( Address & 0x07 )
                		{
                			case 1:
								if( ( value > 0x7F ) && ( value < 0xA0 ) )
								{
									value = 0;
								}
                				DataArray[1] = value;
								Source = ( ( DataArray[1] << 8 ) | DataArray[2] );
								break;
                			case 2:
                				DataArray[2] = (byte)( value & 0xF0 );
                				Source = ( ( DataArray[1] << 8 ) | DataArray[2] );
								break;
                			case 3:
                				DataArray[3] = (byte)( value & 0x1F );
								Destination = ( 0x8000 | ( ( DataArray[3] << 8 ) | DataArray[4] ) );
								break;
                			case 4:
                				DataArray[4] = (byte)( value & 0xF0 );
                				Destination = ( 0x8000 | ( ( DataArray[3] << 8 ) | DataArray[4] ) );
								break;
							case 5:
								Count = ( 16 + ( ( value & 0x7F ) << 4 ) );
								if( Enabled )
								{
									if( ( value & 0x80 ) != 0 )
									{
										DataArray[5] = (byte)( value & 0x7F );
									}
									else
									{
										DataArray[5] = 0xFF;
										Enabled = false;
									}
								}
								else
								{
									if( ( value & 0x80 ) != 0 )
									{
										Enabled = true;
										DataArray[5] = (byte)( value & 0x7F );
										if( LCDC.ModeFlag == 0 )
										{
											Transfer();
										}
									}
									else
									{
										//Console.Processor.Cycles += ( 231 + ( ( value & 0x7F ) << 3 ) );
										HaltFlag = true;
										Console.Memory.Copy( Source, Destination, Count );
										Destination += Count;
										Source += Count;
										DataArray[1] = (byte)( ( Source >> 8 ) & 0xFF );
										DataArray[2] = (byte)( Source & 0xF0 );
										DataArray[3] = (byte)( ( Destination & 0x1FFF ) >> 8 );
										DataArray[4] = (byte)( Destination & 0xF0 );
									}
								}
								break;
							default:
                                break;
                		}
                	}
                }
                
                public void Transfer()
                {
					Console.Memory.Copy( Source, Destination, 0x10 );
					Destination += 0x10;
					Source += 0x10;
					DataArray[2] += 0x10;
					if( DataArray[2] == 0 )
					{
						DataArray[1]++;
					}
					DataArray[4] += 0x10;
					if( DataArray[4] == 0 )
					{
						DataArray[3]++;
					}
					Count -= 0x10;
					DataArray[5]--;
					if( DataArray[5] == 0xFF )
					{
						Enabled = false;
					}
                	return;
                }

			}
			
		}
		
	}
	
}






