using System;
using Collections = System.Collections.Generic;
using GameKnight.Core;

namespace GameKnight
{
	
	namespace GameBoy
	{
		
		public partial class VideoProcessor : IVideoProcessor
		{

			private enum States : int
			{
				HBlank = 0,
				VBlank = 1,
				OAM = 2,
				RAM = 3,
			}
            
			private readonly IAttributeTable[] AttributeTableArray = null;

			private static readonly uint[] ColorArray = new uint[256*256];

			private readonly IConsole Console = null;
				
			private int CurrentFPS = 0;
			
			public int FPS
			{
				get
				{
					return FPS_Property;
				}
				private set
				{
					FPS_Property = value;
				}
			}
			
			private int FPS_Property = 0;
			
			private int Frame = 0;
			
			private int FrameSkip = 0;
	
			private double FrameTime = 0.0;
			
			private readonly Register.IE IE = null;
			
			private static readonly int[] IndexArray = new int[256*256];
            
			private readonly IOptions Options = null;
			
			private readonly Register.HDMA HDMA = null;

			private readonly Register.LCDC LCDC = null;

			private readonly IMap[] MapArray = null;

			private readonly INameTable[] NameTableArray = null;

			private readonly IPalette[] PaletteArray = null;

			private readonly IPatternTable[] PatternTableArray = null;

			private readonly Register.SCR SCR = null;

			public ISize ScreenSize
			{
				get
				{
					return new Size( 160, 144 );
				}
			}
			
            //private static readonly uint[][] SpriteColorArray = new uint[160][];
	         
			private readonly Collections.List<ISprite> SpriteList = new Collections.List<ISprite>();

			private States State
			{
				get
				{
					return (States)LCDC.ModeFlag;
				}
				set
				{
					LCDC.ModeFlag = (int)value;
				}
			}

			private static readonly double[] StateTimingArray = new double[] { 0.00004802, 0.000114, 0.00001931, 0.00004137 };

			//private static readonly double[] StateTimingArray = new double[] { 205.0, 456.0, 80.0, 171.0 };

			/*public Direct3D9.Texture Texture
			{
				get
				{
					return TextureProperty;
				}
				private set
				{
					TextureProperty = value;
				}
			}

			private Direct3D9.Texture TextureProperty = null;*/

			private double Time = 0.0;
			
			//private readonly Timer Timer = new Timer();
			
		/*	private readonly CustomVertex[] VertexArray =
			{
				new CustomVertex( 0.0f, 0.0f, ( 1.0f / 256.0f ) * 160.0f, 0.0f ),
				new CustomVertex( 160.0f, 0.0f, 0.0f, 0.0f ),
				new CustomVertex( 0.0f, 144.0f, ( 1.0f / 256.0f ) * 160.0f, ( 1.0f / 256.0f ) * 144.0f ),
				new CustomVertex( 160.0f, 144.0f, 0.0f, ( 1.0f / 256.0f ) * 144.0f ),
			};*/
							
			private readonly Register.WIN WIN = null;

			public VideoProcessor( IConsole Console )
			{
				this.Console = Console;
				this.Options = Console.Interface.Options;
				if( Options != null )
				{
					this.Options.Get( "VideoProcessor", "FrameSkip", ref FrameSkip );
					this.Options.ValueChanged += OptionsChangedEvent;
				}
				this.IE = Console.GetRegister<Register.IE>();
				this.LCDC = Console.GetRegister<Register.LCDC>();
				this.SCR = Console.GetRegister<Register.SCR>();
				this.WIN = Console.GetRegister<Register.WIN>();
				this.HDMA = Console.GetRegister<Register.HDMA>();
				this.PaletteArray = Console.VideoMemory.GetPalettes();
				this.SpriteList.AddRange( Console.VideoMemory.GetSprites() );
				this.NameTableArray = Console.VideoMemory.GetNameTables();
				this.AttributeTableArray = Console.VideoMemory.GetAttributeTables();
				this.MapArray = Console.VideoMemory.GetMaps();
				this.PatternTableArray = Console.VideoMemory.GetPatternTables();
				return;
			}

			private void DrawBackground()
			{
				if( LCDC.DisplayEnabled )
				{
					if( LCDC.BackgroundEnabled )
					{
						for( int x = 0; x <= 160; x += 8 )
						{
							int Index = ( ( ( ( x >> 3 ) + ( SCR.X >> 3 ) ) & 0x1F ) + ( ( ( ( LCDC.ScanLine + SCR.Y ) >> 3 ) & 0x1F ) * MapArray[LCDC.Map].Size.Width ) );
							MapArray[LCDC.Map].DrawTile( Index, ( x - ( SCR.X & 0x07 ) ), LCDC.ScanLine, ( ( LCDC.ScanLine + SCR.Y ) & 0x07 ), PatternTableArray, IndexArray, ColorArray );
						}
					}
	    			if( ( LCDC.WindowEnabled ) && ( LCDC.ScanLine >= WIN.Y ) )
					{
						for( int x = WIN.X, i = 0; x <= 160; x += 8, i++ )
						{
							int Index = ( ( i & 0x1F ) + ( ( ( ( LCDC.ScanLine - WIN.Y ) >> 3 ) & 0x1F ) * MapArray[LCDC.Window].Size.Width ) );
							MapArray[LCDC.Window].DrawTile( Index, x, LCDC.ScanLine, ( ( LCDC.ScanLine - WIN.Y ) & 0x07 ), PatternTableArray, IndexArray, ColorArray );
						}
					}
				}
				else
				{
					for( int x = 0; x < 160; x++ )
					{
						ColorArray[(LCDC.ScanLine<<8)+x] = PaletteArray[0].ColorArray[0];
					}
				}
				return;
			}

			private void DrawSprites()
			{
				//SpriteList.Sort();
				if( ( LCDC.DisplayEnabled ) && ( LCDC.SpritesEnabled ) )
				{
					for( int i = 0; i < SpriteList.Count; i++ )
					{
						if( ( LCDC.ScanLine >= SpriteList[i].Y ) && ( LCDC.ScanLine < SpriteList[i].Y + LCDC.SpriteHeight ) )
						{
							if( ( SpriteList[i].X > -8 ) && ( SpriteList[i].X < 160 ) )
							{
								SpriteList[i].Draw( LCDC.ScanLine, IndexArray, ColorArray );
							}
						}
					}
				}
				return;
			}
			
			private void OptionsChangedEvent( object Sender, EventArgs Arguments )
			{
				Options.Get( "VideoProcessor", "FrameSkip", ref FrameSkip );
				return;
			}

			public void Reset()
			{
				for( int i = 0; i < ColorArray.Length; i++ )
				{
					ColorArray[i] = 0;
				}
				State = (States)0;
				Time = 0.0;
				return;
			}
			
			public double Run( double TimePassed )
			{
				Time += TimePassed;
				if( Time >= StateTimingArray[LCDC.ModeFlag] )
				{
					Time -= StateTimingArray[LCDC.ModeFlag];
					TimePassed = StateTimingArray[LCDC.ModeFlag];
					switch( State )
					{
						case States.HBlank:
							State = States.OAM;
							break;
						case States.VBlank:
							LCDC.ScanLine++;
							if( LCDC.ScanLine == 0 )
							{
								State = States.HBlank;
							}
							break;
						case States.OAM:
							State = States.RAM;
							break;
						case States.RAM:
							if( Frame >= FrameSkip )
							{
								DrawBackground();
								DrawSprites();
							}
							LCDC.ScanLine++;
							if( LCDC.ScanLine >= 144 )
							{
								CurrentFPS++;
								State = States.VBlank;
								if( Frame >= FrameSkip )
								{
									Frame = 0;
									Console.Interface.Present( ColorArray );
								}
								else
								{
									Frame++;
								}
							}
							else
							{
								State = States.HBlank;	
							}
							if( ( HDMA != null ) && ( HDMA.Enabled ) )
							{
								HDMA.Transfer();
							}
							break;
						default:
							break;
					}		
				}
				FrameTime += TimePassed;
				if( FrameTime >= 1.0 )
				{
					FrameTime -= 1.0;
					FPS = CurrentFPS;
					CurrentFPS = 0;
				}
				return TimePassed;
			}
			
		}
		
	}
	
}
