﻿using System;
using Collections = System.Collections.Generic;
using IO = System.IO;
using GameKnight.Core;

namespace GameKnight.Win32
{

	public sealed class ConfigFile : IOptions
	{
		
		private readonly bool AutoSave = false;
		
		public bool Empty
		{
			get
			{
				return ( ( ValueList.Count <= 1 ) && ( ValueList["root"].Count == 0 ) );
			}
		}
		
		private readonly string Path = "";
		
		public event EventHandler ValueChanged = null;
		
		private readonly Collections.Dictionary<string,Collections.Dictionary<string,string>> ValueList = new Collections.Dictionary<string,Collections.Dictionary<string,string>>();
		
		public ConfigFile( string Path ) : this( Path, false )
		{
			return;
		}
		
		public ConfigFile( string Path, bool AutoSave )
		{
			string[] StringArray = null;
			this.Path = Path;
			this.AutoSave = AutoSave;
			this.ValueList.Add( "root", new Collections.Dictionary<string,string>() );
			try
			{
				StringArray = IO.File.ReadAllLines( Path );
			}
			catch
			{
				return;
			}
			string CurrentKey = "root";
			for( int i = 0; i < StringArray.Length; i++ )
			{
				string CurrentString = StringArray[i].Trim( ' ', '\t' );
				if( ( CurrentString.StartsWith( "[" ) ) && ( CurrentString.EndsWith( "]" ) ) )
				{
					CurrentKey = CurrentString.Substring( 1, CurrentString.Length - 2 );
					if( ValueList.ContainsKey( CurrentKey ) == false )
					{
						this.ValueList.Add( CurrentKey, new Collections.Dictionary<string,string>() );
					}
				}
				else if( CurrentString.Contains( "=" ) )
				{
					string Name = CurrentString.Substring( 0, CurrentString.IndexOf( '=' ) );
					if( Name != "" )
					{
						this.ValueList[CurrentKey].Add( Name, CurrentString.Remove( 0, CurrentString.IndexOf( '=' ) + 1 ) );
					}
				}
			}
			return;
		}
		
		public void Get( string Key, string Name, ref string Value )
		{
			string Result = Read( Key, Name );
			if( Result != "" )
			{
				Value = Read( Key, Name );
			}
			return;
		}
		
		public void Get( string Key, string Name, ref int Value )
		{
			int Result = 0;
			if( int.TryParse( Read( Key, Name ), out Result ) )
			{
				Value = Result;
			}
			return;
		}
		
		public void Get( string Key, string Name, ref double Value )
		{
			double Result = 0;
			if( double.TryParse( Read( Key, Name ), out Result ) )
			{
				Value = Result;
			}
			return;
		}
		
		public void Get<T>( string Key, string Name, ref T Value )
		{
			if( typeof(T).BaseType != typeof(Enum) )
			{
				return;
			}
			object Result = null;
			try
			{
				Result = Enum.Parse( typeof( T ), Read( Key, Name ), true );
			}
			catch
			{
				return;
			}
			Value = (T)Result;
			return;
		}
		
		private string Read( string Key, string Name )
		{
			if( ValueList.ContainsKey( Key ) )
			{
				if( ValueList[Key].ContainsKey( Name ) )
				{
					return ValueList[Key][Name];
				}
			}
			return "";
		}
		
		public bool Save()
		{
			IO.StreamWriter StreamWriter = null;
			try
			{
				StreamWriter = new IO.StreamWriter( IO.File.Open( Path, IO.FileMode.Create, IO.FileAccess.Write ) );
			}
			catch
			{
				return false;
			}
			Collections.Dictionary<string,Collections.Dictionary<string,string>>.Enumerator PrimaryEnumerator = ValueList.GetEnumerator();
			do
			{
				if( PrimaryEnumerator.Current.Value != null )
				{
					if( PrimaryEnumerator.Current.Key != "root" )
					{
						StreamWriter.WriteLine( "[" + PrimaryEnumerator.Current.Key + "]" );
					}
					Collections.Dictionary<string,string>.Enumerator SecondaryEnumerator = PrimaryEnumerator.Current.Value.GetEnumerator();
					do
					{
						if( SecondaryEnumerator.Current.Value != null )
						{
							StreamWriter.WriteLine( SecondaryEnumerator.Current.Key + "=" + SecondaryEnumerator.Current.Value );
						}
					}
					while( SecondaryEnumerator.MoveNext() );
					StreamWriter.WriteLine();
				}
			}
			while( PrimaryEnumerator.MoveNext() );
			StreamWriter.Close();
			return true;
		}
		
		public void Set( string Key, string Name, string Value )
		{
			Write( Key, Name, Value );
			return;
		}
		
		public void Set( string Key, string Name, int Value )
		{
			Write( Key, Name, Value );
			return;
		}
		
		public void Set<T>( string Key, string Name, T Value )
		{
			if( typeof(T).BaseType != typeof(Enum) )
			{
				return;
			}
			Write( Key, Name, Value );
			return;
		}
		
		private void Write( string Key, string Name, object Value )
		{
			if( ( Key == null ) || ( Name == null ) || ( Value == null ) )
			{
				return;
			}
			if( ValueList.ContainsKey( Key ) == false )
			{
				if( ( Key == "root" ) || ( Key == "" ) )
				{
					return;
				}
				ValueList.Add( Key, new Collections.Dictionary<string,string>() );
			}
			if( ValueList[Key].ContainsKey( Name ) )
			{
				ValueList[Key][Name] = Value.ToString();
			}
			else if( Name != "" )
			{
				ValueList[Key].Add( Name, Value.ToString() );
			}
			if( AutoSave )
			{
				Save();
			}
			if( ValueChanged != null )
			{
				ValueChanged( this, EventArgs.Empty );
			}
			return;
		}
		
	}
	
}
